# standard imports
from typing import TypedDict, List

# external imports

# project imports

class Environment(TypedDict):
    connection_string: str
    server: str
    database: str
    username: str
    password: str

class Environments(TypedDict):
    environment: Environment

class Manifest(TypedDict):
    db_engine: str
    environments: dict
    migrations: List[str]



# standard imports
from email.policy import default
import platform
from argparse import ArgumentParser
import pathlib as pl
from pydoc import cli
from typing import List
import getpass

# external imports
import click

# project imports
import wigeon
from wigeon.packages import Package
from wigeon.levels import get_level


###############
### Module-level variables
###############
user = getpass.getuser()

###############
### Call back functions
###############
def print_version(
    ctx: click.Context,
    param: click.Parameter,
    value: bool
) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo(
        f"Running wigeon version {wigeon.__version__} with {platform.python_implementation()} {platform.python_version()} on {platform.system()}"
    )
    ctx.exit()

def nerd(
    ctx: click.Context,
    param: click.Parameter,
    value: bool
) -> None:
    """
    For the funnies.
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo("wigeon recalls... ", nl=False)
    click.echo(click.style(get_level(),  bg='blue', fg='white'))
    ctx.exit()


###############
### cli base
###############
@click.group("wigeon")
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Display the wigeon version and exit.",
)
@click.option(
    "--nerd",
    is_flag=True,
    callback=nerd,
    expose_value=False,
    is_eager=True,
    help="wigeon nerds"
)
def app():
    # present only for sub-commands
    pass


###############
### create commands
###############
@app.group("create")
def create():
     # present only for sub-commands
    pass

@create.command("package")
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of the package you want to create"
)
@click.option(
    "-d",
    "--dbtype",
    type=str,
    help="The db type this package will connect to (either sqlite or mssql currently)"
)
@click.option(
    "-e",
    "--environments",
    type=str,
    default="local,dev,qa,prod",
    help="A comma delmimted string of the enviornments in your db deployment "
)
def create_package(
    name: str,
    dbtype: str,
    environments: str
):
    """
    createpackage initializes a package of migrations in the current
    environment. A package is linked directly to a database type and the
    deployment environments in your ci/cd pipeline.

    dbtype either sqlite, mssql, or postgres
    """
    click.echo(f"Creating {name} package")
    click.echo(f"{name}'s Database type is {dbtype}")
    click.echo(f"{name}'s environments include {environments.split(',')}")
    # check if package exists
    package = Package(packagename=name)
    package.exists(
        raise_error_on_exists=True,
        raise_error_on_not_exist=False
    )

    # initialize package folder
    package.create(
        env_list=environments.split(","),
        db_engine=dbtype
    )

    click.echo(click.style(f"Successfully created package: {name}!", fg="green"))

@create.command("migration")
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of the migration you want to create"
)
@click.option(
    "-p",
    "--package",
    type=str,
    help="The name of the package you want to add a migration to"
)
@click.option(
    "-b",
    "--build",
    type=str,
    default=None,
    help="The build version you want to apply to the migration (use #.#.# format)"
)
def create_migration(
    name: str,
    package: str,
    build: str    
):
    """
    createmigration initializes a .sql module for a migration
    """
    click.echo(f"Creating {name} in {package} package...")
    # check if package exists
    package = Package(packagename=package)
    package.exists()
    # find latest migration number
    current_migrations = package.list_local_migrations()
    current_migr_num = package.find_current_migration(migration_list=current_migrations)
    click.echo(f"Current migration is: {current_migr_num}")
    # create migration
    package.add_migration(
        current_migration=current_migr_num,
        migration_name=name,
        builds=[build] # TODO enable multiple build support at later date
    )
    click.echo(click.style(f"Successfully created {current_migr_num}-{name}.sql!!", fg="green"))


###############
### show commands
###############
@app.command("show")
@click.option(
    "-n",
    "--name",
    default=None,
    type=str,
    help="The name of the package you want to check"
)
@click.option(
    "-p",
    "--packages",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="Shows the package in your workspace"
)
@click.option(
    "-m",
    "--migrations",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="Shows the migrations in your manifest/workspace"
)
@click.option(
    "-b",
    "--builds",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="Shows the builds in your manifest"
)
def show(
    name: str,
    packages: bool,
    migrations: bool,
    builds: bool
):

    if packages:
        raise NotImplementedError("Show packages not yet implemented")
    
    if builds:
        raise NotImplementedError("Show builds not yet implemented")
    
    if migrations and not name:
        raise ValueError("You must provide a package name to show it's migrations (-n or --name)")

    # list migration flag option
    if migrations:
        # check if package exists and read data
        package = Package(packagename=name)
        package.exists()
        package.read_manifest()

        click.echo(f"Found following migrations for {name}:")
        current_migrations = package.list_local_migrations()
        for m in sorted(current_migrations):
            click.echo(f"    {m.name}")
        current_migr = package.find_current_migration(migration_list=current_migrations)
        click.echo(click.style(f"Current migration would be: {current_migr}", fg="yellow"))
        return

###############
### config commands
###############
@app.command("config")
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of the package you want to config"
)
@click.option(
    "-s",
    "--server",
    type=str,
    default=None,
    help="The server address for your db (host:port)"
)
@click.option(
    "-d",
    "--database",
    type=str,
    default=None,
    help="The name of the database"
)
@click.option(
    "-dt",
    "--dbtype",
    type=str,
    default=None,
    help="The db type this package will connect to (either sqlite or mssql currently)"
)
@click.option(
    "-u",
    "--username",
    type=str,
    default=None,
    help="The db server username"
)
@click.option(
    "-p",
    "--password",
    type=str,
    default=None,
    help="The db server password"
)
@click.option(
    "-c",
    "--conn_string",
    type=str,
    default=None,
    help="The full connection string"
)
@click.option(
    "-e",
    "--environment",
    type=str,
    default=None,
    help="The name of the environment config that you are updating"
)
@click.option(
    "-f",
    "--force_nulls",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="If param not provided, will overwrite config field with null"
)
def config(
    name:str,
    server: str,
    database: str,
    dbtype: str,
    username: str,
    password: str,
    conn_string: str,
    environment: str,
    force_nulls: bool
):
    """
    Updates configuration of an manifest for a given environment
    """
    # check if package exists and read data
    package = Package(packagename=name)
    package.exists()
    package.read_manifest()
    package.update_config(
        db_engine=dbtype,
        environment=environment,
        connectionstring=conn_string,
        server=server,
        database=database,
        username=username,
        password=password,
        force_nulls=force_nulls
    )
    package.write_manifest()

###############
### migrate commands
###############
@app.command("migrate")
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of the package you want to migrate"
)
@click.option(
    "-s",
    "--server",
    type=str,
    default=None,
    help="The server address for your db (host:port)"
)
@click.option(
    "-b",
    "--buildtag",
    type=str,
    default=None,
    help="The build version you want to run up to (format #.#.#)"
)
@click.option(
    "-d",
    "--database",
    type=str,
    default=None,
    help="The name of the database"
)
@click.option(
    "-u",
    "--username",
    type=str,
    default=None,
    help="The db server username"
)
@click.option(
    "-p",
    "--password",
    type=str,
    default=None,
    help="The db server password"
)
@click.option(
    "-dr",
    "--driver",
    type=str,
    default=None,
    help="The driver to be used to connect (ODBC)"
)
@click.option(
    "-c",
    "--conn_string",
    type=str,
    default=None,
    help="The full connection string"
)
@click.option(
    "-e",
    "--environment",
    type=str,
    default=None,
    help="The name of the environment you are targeting (forces read of env vars)"
)
@click.option(
    "-ct",
    "--connect_test",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="Force use of env variable as named in manifest environments",
)
@click.option(
    "-a",
    "--run_all",
    is_flag=True,
    type=bool,
    default=False,
    expose_value=True,
    is_eager=True,
    help="Run all migrations to target db",
)
def migrate(
    name:str,
    server: str,
    database: str,
    username: str,
    password: str,
    driver: str,
    conn_string: str,
    environment: str,
    connect_test: bool,
    run_all: bool,
    buildtag: str
):
    # check if package exists and read data
    package = Package(packagename=name)
    package.exists()
    package.read_manifest()

    # create connection, return cursor
    cnxn = package.connect(
        server=server,
        database=database,
        username=username,
        password=password,
        driver=driver,
        connectionstring=conn_string,
        environment=environment
    )
    click.echo(click.style(f"Successfully connected to {package.manifest['db_engine']} database!!!!", fg="green"))

    # connect_test flag option
    if connect_test:
        cnxn.close()
        return
    
    # MAIN migration

    # initialize changelog table if not exists and add columns
    # change_id, migration_date, applied_by(username), and migration_name(.sql filename)
    package.init_changelog()

    # find migrations already in target database
    db_migrations = package.list_db_migrations()

    # find migrations in manifest
    # filter to migrations only with certain build tag
    mani_migrations = package.list_manifest_migrations(buildtag=buildtag)
    click.echo(f"Migrations in manifest: {mani_migrations}")


    # find migrations alead in the database
    # duplicate_migrations = [m.name for m in mani_migrations if m.name in db_migrations]
    duplicate_migrations = package.compare_migrations(
        manifest_migrations=mani_migrations,
        db_migrations=db_migrations
    )
    click.echo(f"Migrations already in db: {duplicate_migrations}")
    # remove duplicate migrations from manifest, unless all option is given
    if not run_all:
        mani_migrations = [m for m in mani_migrations if m.name not in db_migrations]

    click.echo(f"Migrations to run: {mani_migrations}")
    if len(mani_migrations) > 0:
        for mig in mani_migrations:
            if mig.name in db_migrations:
                duplicate_migrations.append(mig.name)
                continue
            click.echo(f"Running migration {mig}... ")
            mig.run(
                package_path=package.pack_path,
                cursor=package.cursor,
                user=user,
                db_engine=package.manifest["db_engine"]
            )
        click.echo(click.style(f"Successfully migrated {len(mani_migrations)} migrations", fg="green"))
        package.connection.commit()
        package.connection.close()
    else:
        click.echo("No migrations to migrate, wigeon is flying home...")
    print()

if __name__ == "__main__":
    app()


import random

names = [
    "Reveille",
    "AI Constructs and Cyborgs First!",
    "Flawless Cowboy",
    "Reunion Tour",
    "The Truth and Reconciliation",
    "Into the Belly of the Beast",
    "Shut Up and Get Behind me... Sir",
    "The Silent Cartographer",
    "It's Quiet...",
    "Shafted",
    "I Would Have Been Your Daddy...",
    "Rolling Thunder",
    "If I had a Super Weapon...",
    "Well Enough Alone",
    "The Flood",
    "343 Guilty Spark",
    "The Library",
    "Wait, It Gets Worse!",
    "But I Don't Want to Ride the Elevator!",
    "Fourth Floor: Tools, Guns, Keys to Super Weapons",
    "The Gun Point at the Head of the Universe",
    "Breaking Stuff to Look Tough",
    "The Tunnels Below",
    "Final Run",
    "Under New Management",
    "Upstairs, Downstairs",
    "The Captain",
    "...And the Horse You Rode in on",
    "Light Fuse, Run Away",
    "Warning: Hitchhikers May be Escaping Convicts"
]

def get_level():
    return names[random.randrange(len(names))]


# standard imports
import pathlib as pl
import shutil
import json
from typing import List

# external imports
import pymssql

# project imports
from wigeon.db import Connector, Migration
from wigeon.manifests import Environment, Manifest


class Package(object):
    pack_folder = pl.Path().cwd().joinpath("packages")

    def __init__(
        self,
        packagename:str
    ):
        self.packagename = packagename
        self.pack_path = Package.pack_folder.joinpath(packagename)
        self.manifest = self.read_manifest()
        self.connector = None
        self.connection = None
        self.cursor = None

    def exists(
        self,
        raise_error_on_exists: bool=False,
        raise_error_on_not_exist: bool=True
    ) -> bool:
        """
        Checks if a given package exists. Defaults to error if package is not present.
        """
        package_exists = Package.pack_folder.joinpath(self.packagename).exists()
        if package_exists and raise_error_on_exists:
            raise FileExistsError(
                f"A package with the name {self.packagename} already exists at {Package.pack_folder.joinpath(self.packagename)}."
            )
        elif (not package_exists) and (raise_error_on_not_exist):
            raise FileExistsError(
                f"A package with the name {self.packagename} does not exist at {Package.pack_folder.joinpath(self.packagename)}."
            )
        return package_exists
    
    def create(
        self,
        env_list: list,
        db_engine: str
    ):
        """
        Package.create initializes a package in the root/packages/ directory and
        supplies an __init__.py and manifest.json
        """
        # initialize package folder
        self.pack_path.mkdir(parents=True)
        with open(self.pack_path.joinpath("__init__.py"), "w") as f:
            f.write("# auto generated package initializer")
        
        # initialize package manifest
        manifest_template = {}
        manifest_template["db_engine"] = db_engine
        manifest_template["environments"] = {}
        for e in env_list:
            manifest_template["environments"][e] = {
                "connectionstring": None,
                "server": None,
                "database": None,
                "username": None,
                "password": None
            }
        manifest_template["migrations"] = []

        with open(self.pack_path.joinpath("manifest.json"), "w") as f:
            json.dump(manifest_template, f, indent=4)
    
    def delete(self):
        """
        Package.delete removes the package directory from os
        """
        shutil.rmtree(self.pack_path, ignore_errors=True)
    
    def list_local_migrations(
        self
    ) -> list:
        """
        list_mgrations reads a package and returns a list of all the migrations in the folder
        """
        return [f for f in Package.pack_folder.joinpath(self.packagename).iterdir() if f.suffix == ".sql"]

    def find_current_migration(
        self,
        migration_list: List[pl.Path]
    ) -> str:
        """
        Parses local migration files and finds version number to return the latest as a string.
        
        Assumes migration filename convention of:
        ####-<migration_name>.sql
        """
        # BUG latest migration algorithm is erroring sometimes
        if migration_list == []:
            latest_migration_int = 0
        else:
            # sort migration list to ensure largest number is the latest
            migration_list = sorted(migration_list)
            latest_migration_int = int(
                str(
                    migration_list[-1].name
                )[0:4]
            )
        
        # increment migration guid
        current_migration = latest_migration_int + 1
        # check if migration greater than limit of 4 characters
        if current_migration > 9999:
            raise ValueError(
                f"Current migration number ({current_migration}) cannot exceed 9999, max characters allowed in migration guid is 4"
            )
        # return 4 character migration guid string left padded with zeros
        return str(current_migration).zfill(4)

    def read_manifest(self):
        try:
            with open(self.pack_path.joinpath("manifest.json"), "r") as f:
                self.manifest = json.load(f)
        except FileNotFoundError:
            print(f"Cannot read: {self.pack_path.joinpath('manifest.json')}, package might not exist or is being initialized.")

    def write_manifest(self):
        if not self.manifest:
            raise ValueError(
                f"{self.__class__} {self.packagename}'s manifest not yet read or built. Wigeon will not write nonetype manifest."
            )
        with open(self.pack_path.joinpath("manifest.json"), "w") as f:
            json.dump(self.manifest, f, indent=4)
    
    def list_manifest_migrations(self, buildtag: str=None):
        """
        list_manifest_migrations collects migrations present in the manifest and
        allows for filtering based on buildtag
        """
        if not self.manifest:
            raise ValueError(
                f"{self.__class__} {self.packagename}'s manifest not yet read, built, or no migrations have been added."
            )
        if buildtag: # pragma: no cover
            # TODO implement filtering by build tag
            raise NotImplementedError("Build tag filtering not yet implemented!")
            migrations = {k:v for k,v in self.manifest["migrations"] if buildtag in v["builds"]}
        else:
            migrations = [Migration(**m) for m in self.manifest["migrations"]]
            #migrations = self.manifest["migrations"].copy()
        
        return migrations
    
    def add_migration(
        self,
        current_migration: str,
        migration_name: str,
        builds: List[str]
    ):
        with open(self.pack_path.joinpath(f"{current_migration}-{migration_name}.sql"), "w") as f:
            f.write("-- TODO build migration code")
        self.read_manifest()
        self.manifest["migrations"].append(
            {
                "name": f"{current_migration}-{migration_name}.sql",
                "builds": [b for b in builds]
            }
        )
        self.write_manifest()
    
    def create_connector(self, environment: str=None):
        self.connector = Connector(
            manifest=self.manifest,
            environment=self.manifest["environments"][environment]
        )
    
    def list_db_migrations(self) -> list:
        # find migrations already in target database
        query_migrations_from_changelog = "SELECT migration_name from changelog;"

        self.cursor.execute(query_migrations_from_changelog)
        try:
            db_migrations = [n[0] for n in self.cursor.fetchall()]
        except IndexError as e:
            print("No index 0 in db migrations, changelog table must not be set up in target database")
            exit()
        return db_migrations
    
    def compare_migrations(
        self,
        manifest_migrations: List[Migration],
        db_migrations: List[str]
        ):
        """
        compare_migrations compares migrations in the manifest to migrations already in the database
        and returns a list of duplicate migrations
        """
        # find migrations alead in the database
        duplicate_migrations = [m.name for m in manifest_migrations if m.name in db_migrations]
        return duplicate_migrations

    def init_changelog(self):
        # initialize changelog table if not exists and add columns
        # change_id, migration_date, applied_by(username), and migration_name(.sql filename)
        try:
            self.cursor.execute(self.connector.changeloginit)
        except pymssql._pymssql.OperationalError:
            print("changelog exists")
    
    def connect(
        self,
        server: str=None, # connection variable
        database: str=None, # connection variable
        username: str=None, # connection variable
        password: str=None, # connection variable
        driver: str=None, # connection variable
        connectionstring: str=None, # connection variable,
        environment: Environment=None
    ):
        self.create_connector(environment=environment)
        self.connection = self.connector.connect(
            server=server,
            database=database,
            username=username,
            password=password,
            driver=driver,
            connectionstring=connectionstring
        )
        self.cursor = self.connection.cursor()

        return self.cursor
    
    def update_config(
        self,
        db_engine: str=None,
        environment: str=None,
        connectionstring: str=None,
        server: str=None,
        database: str=None,
        username: str=None,
        password: str=None,
        force_nulls: bool=False
    ):
        if force_nulls:
            self.manifest["db_engine"] = db_engine
            self.manifest["environments"][environment]["connectionstring"] = connectionstring
            self.manifest["environments"][environment]["server"] = server
            self.manifest["environments"][environment]["database"] = database
            self.manifest["environments"][environment]["username"] = username
            self.manifest["environments"][environment]["password"] = password
        else:
            if db_engine:
                self.manifest["db_engine"] = db_engine
            if connectionstring:
                self.manifest["environments"][environment]["connectionstring"] = connectionstring
            if server:
                self.manifest["environments"][environment]["server"] = server
            if database:
                self.manifest["environments"][environment]["database"] = database
            if username:
                self.manifest["environments"][environment]["username"] = username
            if password:
                self.manifest["environments"][environment]["password"] = password
        
        self.write_manifest()


# standard imports
import pathlib as pl
import os
import sqlite3
import datetime
from typing import Union, List

# external imports
# import pyodbc
import pymssql

# project imports
from wigeon.manifests import Environment, Manifest

class Connector(object):

    db_engines = [
        "sqlite",
        "mssql",
        "postgres"
    ]

    def __init__(
        self,
        manifest: Manifest,
        environment: Environment
    ):
        self.db_engine = manifest["db_engine"]
        self.manifest = manifest
        self.environment = environment
        self.cnxn = None
        self.changeloginit = {
            "sqlite": "CREATE TABLE IF NOT EXISTS changelog (change_id INTEGER NOT NULL PRIMARY KEY, migration_date TEXT, migration_name TEXT, applied_by TEXT);",
            "mssql":  """CREATE TABLE changelog (change_id INT NOT NULL IDENTITY PRIMARY KEY, migration_date TEXT, migration_name TEXT, applied_by TEXT);"""
        }[self.db_engine]


    def connect(
        self,
        **kwargs
    ):
        db_engines = {
            "sqlite": self.conn_sqlite,
            "mssql": self.conn_mssql,
            "postgres": self.conn_postgres
        }
       
       # read environment name from Connector and collect envvariable names
       # extract environment variables to kwargs if variables exist
        if self.environment:
            print(f"Connecting to {self.db_engine} via environment variables...", end=" ")
            kwargs = self.environment
            # dictionary comprehension ftwftwftw
            kwargs = {k:os.environ[v] for k,v in kwargs.items() if v}

         # run connection method based on db_engine for package
        db_engines[self.db_engine](**kwargs)
        return self.cnxn
        

    def conn_sqlite(
        self,
        **kwargs
    ) -> sqlite3.Connection:
        """
        Connect to a sqlite database and return connection
        """
        if kwargs["connectionstring"]:
            self.cnxn = sqlite3.connect(kwargs["connectionstring"])
            return self.cnxn
        else:
            raise ValueError("sqlite connection requires connectionstring argument only (the filepath)")
    
    def conn_postgres(self, **kwargs):
        raise NotImplementedError("conn_postgres is not yet implemented!")
    
    def conn_mssql(
        self,
        **kwargs
    ) -> pymssql.Connection:
        """
        Connect to a sql server database and return connection
        
        server requires host and port example: "0.0.0.0:1433"
        """
        try:
            kwargs["connectionstring"]
            raise NotImplementedError("connectionstring not yet supported for mssql connection")
        except:
            pass
        self.cnxn = pymssql.connect(
            server=kwargs["server"],
            user=kwargs["username"],
            password=kwargs["password"],
            database=kwargs["database"]
        )
        return self.cnxn

class Migration(object):

    def __init__(
        self,
        name: str,
        builds: List[str]
    ):
        self.name = name
        self.builds = builds
    
    def __str__(self):
        return f"name: {self.name}, builds: {self.builds}"
    
    def __repr__(self):
        return f"name: {self.name}, builds: {self.builds}"

    def run(
        self,
        package_path: pl.Path,
        cursor: Union[sqlite3.Cursor, pymssql.Cursor],
        user: str,
        db_engine: str
    ):
        with open(package_path.joinpath(self.name), "r") as f:
            query = f.read()
        
        try:
            cursor.execute(query)
        except Exception as e:
            if db_engine == "sqlite":
                cursor.execute("ROLLBACK")
                raise e
            if db_engine == "mssql":
                raise e
                

        migration_date = datetime.datetime.now().strftime("%Y%m%d-%H%M")

        if db_engine == "sqlite":
            cursor.execute(
                "INSERT INTO changelog (migration_date, migration_name, applied_by) VALUES(:migration_date, :migration_name, :applied_by)",
                {
                    "migration_date": migration_date,
                    "migration_name": self.name,
                    "applied_by": user
                }
            )
        if db_engine == "mssql":
            cursor.execute(
                "INSERT INTO changelog (migration_date, migration_name, applied_by) VALUES (%s, %s, %s)",
                (migration_date, self.name, user)
            )